#!/bin/sh -e

# Run a CAIA competition, but many matches in parallel using xjobs.
# If it doesn't work: apt-get install xjobs
# Keep the competition-stubs and -results directories empty for best results.

# Author: Wilmer van der Gaast <wilmer@gaast.net>

BASE=$HOME/caia/less
STUBS=$BASE/competition-stubs
RES=$BASE/competition-results

# Number of matches to play per pair of players.
NUM_PER_PAIR=2
# Number of miliseconds to allow per player per game.
PLAYER_TIME_LIMIT=30000

DATE_FMT="%Y-%m-%d %H:%M:%S"

if [ -z "$4" ]; then
	echo "Must pass participants as arguments. (At least four players needed.)"
	echo "Example: $0 mysubmission player2 player3 player4"
	exit 1
fi

players="$*"

for player in $players; do
	if [ ! -x $player ]; then
		echo "$player doesn't appear to be a valid executable"
		exit 1
	fi
	if [ -n "$(echo $player | tr -d _A-Za-z0-9)" ]; then
		echo "Executable name $player contains invalid characters."
		echo "It should be in current working directory."
		exit 1
	fi
done

# Don't count the number of logical cores/threads. Although a competition
# finishes slightly faster when using all (hyper)threads, it's only a 10-20%
# improvement instead of the ~100% you'd expect. You just get threads tied
# up waiting for execution units very often. Nice when testing, but it ruins
# the time limits/etc.
if [ -z "$num_cores" ] && [ -e /proc/cpuinfo ]; then
	num_cores=$(grep ^'core id\b' /proc/cpuinfo | sort | uniq | wc -l)
fi

if [ -z "$num_cores" ] || [ "$num_cores" -lt "1" ]; then
	# OS X, src:
	# http://stackoverflow.com/questions/1715580/how-to-discover-number-of-cores-on-mac-os-x
	# http://www.opensource.apple.com/source/xnu/xnu-792.13.8/libkern/libkern/sysctl.h
	num_cores=$(sysctl -n hw.physicalcpu || echo 0)
fi

if [ -z "$num_cores" ] || [ "$num_cores" -lt "1" ]; then
	num_cores=2
	echo "Couldn't figure out number of cores, will guess $num_cores."
else
	echo "Number of cores (w/o Hyper-Threading): $num_cores."
fi

# For good hygiene, you should rm -rf these two dirs first, but I'm always
# very very wary of shellscripts using that command.
mkdir -p $STUBS $RES
rm -rf $STUBS/*.sh

OUT=/proc/$$/fd/1
if [ ! -e "$OUT" ]; then
	# On OS X, we'll have to use a different trick.
	OUT=$BASE/competition-logs.fifo
	[ -e "$OUT" ] || mkfifo "$OUT"
	tail -F "$OUT" &
	
	# I hope this cleanup will be solid..
	tailpid=$!
	trap "kill $tailpid" EXIT
fi

scorefiles=""
# Write a bunch of shellscripts that will work as caiaio clients, instructing
# it to do one match and write the results to the right logs.
for p1 in $players; do
	for p2 in $players; do
		if [ "$p1" = "$p2" ]; then
			continue
		fi
		for p3 in $players; do
			if [ "$p1" = "$p3" -o "$p2" = "$p3" ]; then
				continue
			fi
			for p4 in $players; do
				if [ "$p1" = "$p4" -o "$p2" = "$p4" -o "$p3" = "$p4" ]; then
					continue
				fi
				scores=$RES/$p1-$p2.scores
				scorefiles="$scorefiles $scores"
				if [ -e "$scores" ] && [ "$scores" -nt "$p1" -a "$scores" -nt "$p2" -a "$scores" -nt "$p3" -a "$scores" -nt "$p4" ]; then
					echo "Skipping $p1 - $p2 - $p3 - $p4 since binaries didn't change."
					continue
				fi
				rm -f $scores
				for i in $(seq 1 $NUM_PER_PAIR); do
					f=$STUBS/$p1-$p2-$p3-$p4.$(printf %03d $i).sh
					cat << EOF2 > $f
#!/bin/sh

cat<<EOF
I number_players 4
I player 1 $p1 $PLAYER_TIME_LIMIT $BASE/playerlogs/$p1.in.$p1-$p2-$p3-$p4.$i.log
I start 1
I player 2 $p2 $PLAYER_TIME_LIMIT $BASE/playerlogs/$p2.in.$p1-$p2-$p3-$p4.$i.log
I start 2
I player 3 $p3 $PLAYER_TIME_LIMIT $BASE/playerlogs/$p3.in.$p1-$p2-$p3-$p4.$i.log
I start 3
I player 4 $p4 $PLAYER_TIME_LIMIT $BASE/playerlogs/$p4.in.$p1-$p2-$p3-$p4.$i.log
I start 4
I referee $BASE/bin/referee $BASE/refereelogs/$p1-$p2-$p3-$p4.$i.log reflog.log
EOF
read discard p1s p2s p3s p4s text
if [ \$p1s -gt \$p2s ]; then
	win=$p1
elif [ \$p2s -gt \$p1s ]; then
	win=$p2
else
	win=nobody
fi
scores=\$(printf "\$p1s $p1\n\$p2s $p2\n\$p3s $p3\n\$p4s $p4\n" | sort -rn | tr \\n ,)
echo \$(date +"$DATE_FMT") \$scores >> $RES/$p1-$p2-$p3-$p4.res
echo \$(date +"$DATE_FMT") match $i scores: \$scores >> $OUT
echo $p1 \$p1s >> $scores
echo $p2 \$p2s >> $scores
echo $p3 \$p3s >> $scores
echo $p4 \$p4s >> $scores
cat <<EOF
I kill_referee
I kill 1
I kill 2
I kill 3
I kill 4
I stop_caiaio
EOF
sleep 1
EOF2
					chmod 755 $f
				done
			done
		done
	done
done

if [ -n "$(ls $STUBS/*.sh 2> /dev/null)" ]; then # OSX-- xargs doesn't have -r
	# Have xargs run caiaio with all the stub clients created above. The
	# little-known flag -P tells it to run stuff in parallel.  :-D
	ls $STUBS/*.sh | xargs -P$num_cores -n1 -- $BASE/bin/caiaio -m
else
	echo "No new matches to play."
fi

# Horrible shell hack to sum up all the scores and sort them.
for p in $players; do
	printf "%6d  %-32s\n" $(($(grep -w ^$p $scorefiles | cut -d\  -f2 | xargs | tr \  +))) $p
done | sort -rn
